#define FASTLED_INTERRUPT_RETRY_COUNT 1
#define FASTLED_ALLOW_INTERRUPTS 1
#define USE_OCTOWS2811

#include <OctoWS2811.h>
#include <MIDI.h>
#include <FastLED.h>
#include <SD.h>


// DEFIIIIINEEEE

#define NUM_LEDS_PER_STRIP 256
#define NUM_STRIPS 8
#define NUM_PIXELS 2048
#define BRIGHTNESS 32
#define UPDATE_TIME 22 //ms between each LED update


#define RED    0xFF0000
#define GREEN0  0x008000
#define GREEN  0x00FF00
#define GREEN2  0x11FF11
#define GREEN3  0x22EE22
#define GREEN4  0x33DD33
#define BLUE   0x0000FF
#define BLUE2  0x1111EE
#define BLUE3  0x2222DD
#define BLUE4  0x3333CC
#define BLUE5  0x4444BB
#define YELLOW 0xFFFF00
#define PINK   0xFF1088
#define ORANGE 0xE05800
#define WHITE  0xFFFFFF
#define BLACK  0x000000

#define COLOSC1 0xFF0000
#define COLOSC2 0xD08000
#define COLOSC3 0xB0FF00
#define COLFIL1 0x00BFFF
#define COLAENV 0x9400D3
#define COLMENV 0xFF1493
#define COLLFO1 0x00FF00
#define COLLFO2 0xFFFF00
#define COLEQ1  0x00FFFF
#define COLEQ2  0x30A030
#define COLMFX1 0xFF1493
#define COLMFX2 0x00BFFF


//VARIABLES

CRGB leds[NUM_PIXELS];
CRGBPalette16 VUPalette;
CRGB palettes[18][4];
CRGB modcolor;
CRGB modcolor2;
CRGB modcolor3;

unsigned int previousMillis;         //holds the previous millis
unsigned int fpsMillis;
unsigned int prevMil;
unsigned int saveMillis;

int fpsCount = 0;
int fpsLoop = 0;
int saveCount = 0;

int ringMatrix[3][17];

int shift1 = 0;
int cc[127];
int nrpn[512];
int runner = 0;
int cyclecolor[4];
int times = 10;
int cyclecounter = 0;
int moddot[32];
int moddotnrpn[32];
int shiftdot[32];
int pi;
int pj;

int bpmGlobal = 100;
int bpmGlobal256 = 0;
int bpmGlobal27 = 0;
int bpmGlobalTrigger = 0;
int periodGlobal = 60000 / bpmGlobal;

bool pickedMode = 0;
int saverMode;
int saveTime = 8000;   //800-23sec @30 . // . 1000 = 14sec @UPDATE_TIME 14 hmmmmm.... 14  = 14
int saveChangeTime = 15000;

File presetFile;
const int chipSelect = BUILTIN_SDCARD;




//MIDI_CREATE_DEFAULT_INSTANCE();
MIDI_CREATE_INSTANCE(HardwareSerial, Serial2, MIDI); // teensy pin 8 & 9

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


void setup() {
  randomSeed(analogRead(0));
  delay(300);
  LEDS.addLeds<OCTOWS2811>(leds, NUM_LEDS_PER_STRIP);
  LEDS.setBrightness(BRIGHTNESS);
  clearLEDRingAll();
  setModdot();
  setShiftdot();
  setPalettes();
  setVUPalette();
  setcyclecolor();
  setRingMatrix();
  MIDI.begin();
  Serial.begin(9600);
  Serial.println("RGB LED RING CONTROL v5");
  saveMillis = millis();


  Serial.print("Initializing SD card...");

  if (!SD.begin(chipSelect)) {
    Serial.println("initialization failed!");
    return;
  }
  Serial.println("initialization done.");
}



void loop() {

  /////// READ NEW MIDI DATA, UPDATE CC ARRAY
  midiParse();

  /////// run this loop every "times" milliseconds
  if (millis() >= previousMillis + UPDATE_TIME) {

    previousMillis = millis();
    fpsCount++;
    saveCount++;
    if (runner < NUM_PIXELS - 1) runner++; else runner = 0;

    fpsCounter();

    updateLEDRings();


    /////// OH WAIT MAYBE START SCREENSAVER?
    if (saveCount > saveTime) screensaver();

    /////// UPDATE LEDs AND GO SMOKE A SPLIFF
    LEDS.show();

    /////// GET READY FOR NEXT ROUND
    clearLEDRingAll();  // delete old LED values

  }
  fpsLoop++;
}


////////////////////////////////////////////////////////////////////////////////////
/////////////////////////MIDI FUNCTIONS///////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////


void nrpnParse(void) {
  nrpn[cc[98] + (cc[99] << 7)] = cc[38] + (cc[6] << 7);
  int k = cc[98] + (cc[99] << 7);
  Serial.print("NRPN #"); Serial.print(k); Serial.print(" -  Value "); Serial.println(nrpn[k]);
}


void midiParse() {

  if (MIDI.read())
  {
    //Serial.println("midi");
    saveCount = 0;
    pickedMode = 0;

    if (MIDI.getType() == midi::ControlChange) {
      leds[2] = BLUE; //debug midi monitor ring
      int numtemp = MIDI.getData1();
      int valtemp = MIDI.getData2();
      cc[numtemp] = valtemp;
      Serial.print(numtemp); Serial.print(" . val"); Serial.println(valtemp);
      if (numtemp == 98) {
        int nr = cc[98] + (cc[99] << 7);
        Serial.print("nrpn nr ");
        Serial.print(nr);
        Serial.print(": ");
        Serial.println(nrpn[nr]);
      }
      else if (numtemp == 38) {
        leds[3] = ORANGE;
        nrpnParse();
      }
    }
    else if (MIDI.getType() == midi::SystemExclusive) {
      leds[4] = GREEN;
      if (MIDI.getData1() == 43) {
        if (MIDI.getData2() > 0) shift1 = 1;
      }
    }
    else if (MIDI.getType() == midi::NoteOn) {
      leds[5] = RED;
    }
    if (cc[105] == 1) loadPreset(nrpn[151]);
    else if (cc[105] == 2) savePreset(nrpn[151]);
  }

}



void screensaver() {
  if (!pickedMode) {
    saveMillis = millis();
    saveCount = saveTime + 2;
    saverMode = random(10);
    pickedMode = 1;
    Serial.println(saverMode);
  }
  //clearLEDRingAll();
  //saverMode = 2;
  if (saverMode == 0) saverHueSnake(2);
  else if (saverMode == 1) {
    delay(0);
    saverNightriderFine(1, ((saveCount >> 3) % 15), RED);
  }
  else if (saverMode == 2) saverNightriderCoarse(1, CHSV((saveCount >> 2) % 255, 255, 255));
  else if (saverMode == 3) saverVSwipeAll(1);
  else if (saverMode == 4) saverHSwipeAll(1);
  else if (saverMode == 5) saverAddRandom(0);
  else if (saverMode == 6) saverAddRandomBWFade(10);
  else if (saverMode == 7) {
    delay(1);
    saverNightriderFineH(1, 3, CHSV(saveCount >> 3, 255, 255));
    saverNightriderFine(3, 3, GREEN);
    saverNightriderFine(1, 8, RED);
  }
  else if (saverMode == 8) saverSnowfall(7);
  else if (saverMode == 9) saverHWave(4, 8, CHSV((saveCount >> 4) % 255, 255, 255));
  else if (saverMode == 10) saverNightriderCoarseFade(2, RED);



  if (millis() > (saveMillis + saveChangeTime)) {

    Serial.println("save");
    pickedMode = 0;
  }
}

void fpsCounter() {
  if (previousMillis >= fpsMillis + 8000) {
    fpsMillis = previousMillis;
    Serial.print(fpsCount >> 3); Serial.print(" FPS    -    "); Serial.print(fpsLoop >> 3); Serial.println(" loop() calls per second");
    fpsCount = 0; fpsLoop = 0;
  }
}


void loadPreset(int nr) {
  Serial.println("load");
  char fname[32];
  strcpy(&fname[0], "P000.B1A");
  int i = nr;
  int i0 = i / 10;
  fname[3] = '0' + i - 10 * i0;
  i = i0; i0 = i / 10;
  fname[2] = '0' + i - 10 * i0;
  i = i0; i0 = i / 10;
  fname[1] = '0' + i - 10 * i0;
  presetFile = SD.open(fname);
  if (presetFile) {
    Serial.print("file: ");
    Serial.println(fname);
    presetFile.read((const uint8_t *)&nrpn, sizeof(nrpn));
  }
  presetFile.close();
}

void savePreset(int nr) {
  Serial.println("save");
  char fname[32];
  strcpy(&fname[0], "P000.B1A");

  int i = nr;
  int i0 = i / 10;
  fname[3] = '0' + i - 10 * i0;
  i = i0; i0 = i / 10;
  fname[2] = '0' + i - 10 * i0;
  i = i0; i0 = i / 10;
  fname[1] = '0' + i - 10 * i0;

  if (SD.exists(fname)) SD.remove(fname);
  presetFile = SD.open(fname, FILE_WRITE);

  presetFile.write((const uint8_t *)&nrpn, sizeof(nrpn));
  presetFile.close();
}



////////////////////////////////////////////////////////////////////////////////////
/////////////////////////TEMPO FUNCTIONS///////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////

void bpmCounter(void) {
  if ( (millis() - prevMil) > (bpmGlobal256 * 236 / bpmGlobal) ) {
    bpmGlobal256++;
  }
  if ( (millis() - prevMil) > (bpmGlobal27 * 2315 / bpmGlobal)) {
    bpmGlobal27++;
  }


  Serial.print(bpmGlobal); Serial.print("bpm . "); Serial.print(periodGlobal); Serial.print(" PG .  - ");
  Serial.print(bpmGlobal256); Serial.print(" bpm256 .  - "); Serial.print(bpmGlobal27); Serial.println(" bpm27 .  - ");


  if (millis() > (prevMil + periodGlobal)) {
    prevMil = millis();
    bpmGlobal256 = 0;
    bpmGlobal27 = 0;
    bpmGlobalTrigger = !bpmGlobalTrigger;
  }
}

////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////M/////M/////A/////PPP//////////
/////////////////////////RING MAPPING////////////////MM///MM////A/A////P//P/////////
/////////////////////////////////////////////////////M/M/M/M///A///A///PPP//////////
/////////////////////////////////////////////////////M//M//M//AAAAAAA//P////////////
/////////////////////////////////////////////////////M/ ///M/A///////A/P////////////
////////////////////////////////////////////////////////////////////////////////////


void updateLEDRings() {
  delay(1);
  /////// SET LEDS ARRAY TO NEW VALUES FROM CC ARRAY
  //Serial.println(nrpn[150]);
  if (nrpn[150] == 0) {

    // PRESET
    setLEDPulse(31, 100, GREEN);
    setLEDRingLine(2, 26, CHSV((runner % 1024) / 4 , 255, 255));


    // LFO 1
    setLEDRingSegments(5, nrpn[50], 7, COLLFO1);
    setLEDRingDot(6, map(nrpn[51], 0, 4095, 0, 26), COLLFO1);



    //OSC 1
    setLEDRingDot(9, cc[70] % 27, COLOSC1);
    setLEDRingPitch(10, nrpn[11] - 32, COLOSC1, COLOSC1);
    setLEDRingLine(11, map(nrpn[13], 0, 4095, 0, 26), COLOSC1);
    setLEDRingLine(12, map(nrpn[14], 0, 4095, 0, 26), COLOSC1);
    //Serial.println(nrpn[80]);
    //OSC 2
    setLEDRingDot(17, cc[71] % 27, COLOSC2);
    setLEDRingPitch(18, nrpn[21] - 32, COLOSC2, COLOSC2);
    setLEDRingLine(19, map(nrpn[23], 0, 4095, 0, 26), COLOSC2);
    setLEDRingLine(20, map(nrpn[24], 0, 4095, 0, 26), COLOSC2);
    //OSC3
    setLEDRingDot(25, cc[72] % 27, COLOSC3);
    setLEDRingPitch(26, nrpn[31] - 32, COLOSC3, COLOSC3);
    setLEDRingLine(27, map(nrpn[33], 0, 4095, 0, 26), COLOSC3);
    setLEDRingLine(28, map(nrpn[34], 0, 4095, 0, 26), COLOSC3);
    //MIXER
    setLEDRingVULine(29,  map(nrpn[80], 0, 4095, 0, 26));
    setLEDRingVULine(30,  map(nrpn[81], 0, 4095, 0, 26));
    setLEDRingVULine(31,  map(nrpn[82], 0, 4095, 0, 26));

    //FILTER
    setLEDRingLine(33, map(nrpn[64], 0, 4095, 0, 26), COLFIL1);
    setLEDRingPan(34, map(nrpn[66], 0, 4095, -13, 13), COLFIL1, YELLOW, COLFIL1);
    //setLEDRingFadeIn(35, map(nrpn[60], 0, 4095, 0, 26), ORANGE);
    //setLEDRingFadeOut(36, map(nrpn[61], 0, 4095, 0, 26), PINK);
    // F ADSR
    setLEDRingFadeIn(41, map(nrpn[60], 0, 4095, 0, 26), COLFIL1);
    setLEDRingFadeOut(42, map(nrpn[61], 0, 4095, 0, 26), COLFIL1);
    setLEDRingDot(43, map(nrpn[62], 0, 4095, 0, 26), COLFIL1);
    setLEDRingFadeOut(44, map(nrpn[63], 0, 4095, 0, 26), COLFIL1);

    // A ADSR
    setLEDRingFadeIn(49, map(nrpn[70], 0, 4095, 0, 26), COLAENV);
    setLEDRingFadeOut(50, map(nrpn[71], 0, 4095, 0, 26), COLAENV);
    setLEDRingDot(51, map(nrpn[72], 0, 4095, 0, 26), COLAENV);
    setLEDRingFadeOut(52, map(nrpn[73], 0, 4095, 0, 26), COLAENV);

    // DRIVES

    setLEDRingVULine(57, map(nrpn[90], 0, 4095, 0, 26));
    setLEDRingVULine(58, map(nrpn[91], 0, 4095, 0, 26));

    // EQ1
    setLEDRingDot(59, map(nrpn[110], 0, 4095, 0, 26), COLEQ1);
    setLEDRingPan(60, map(nrpn[111], 0, 4095, -13, 13), COLEQ1, COLEQ1, COLEQ1);

    // MFX 1
    setLEDRingSegments(61, nrpn[100], 15, COLMFX1);
    setLEDRingDot(62, map(nrpn[101], 0, 4095, 0, 26), COLMFX1);
    setLEDRingDot(63, map(nrpn[102], 0, 4095, 0, 26), COLMFX1);
    setLEDRingDot(64, map(nrpn[103], 0, 4095, 0, 26), COLMFX1);

    for (int j = 0; j < 9; j++) {
      for (int i = 0; i < 31; i++) {
        int l = nrpn[201 + (10 * moddotnrpn[i]) + j];
        if (l > 0)
          leds[(moddot[i] * 32) - 1].addToRGB((l >> 6) + 20);
      }
    }



  }

  //////////////// SHIFT 1 (print shift)


  else if (nrpn[150] == 1) {

    // LFO 1
    setLEDRingSegments(5, nrpn[55], 7, COLLFO2);
    setLEDRingDot(6, map(nrpn[56], 0, 4095, 0, 26), COLLFO2);

    //OSC 1 clip//fine
    //setLEDRingVULine(9, map(nrpn[15], 0, 4095, 0, 26));
    setLEDRingPan(10, map(nrpn[12], 0, 4095, -13, 13), COLOSC1, COLOSC1, BLACK);
    //setLEDRingLine(11, map(nrpn[14], 0, 4095, 0, 26), 0x160800);
    //setLEDRingLine(12, map(nrpn[0], 0, 4095, 0, 26), 0x160800);
    //OSC 2 clip/fine/par2/vel
    //setLEDRingVULine(17, map(nrpn[25], 0, 4095, 0, 26));
    setLEDRingPan(18, map(nrpn[22], 0, 4095, -13, 13), COLOSC1, COLOSC1, BLACK);
    //setLEDRingLine(19, map(nrpn[24], 0, 4095, 0, 26), 0x160800);
    //setLEDRingLine(20, map(nrpn[0], 0, 4095, 0, 26), 0x160800);
    //OSC 3
    //setLEDRingVULine(25, map(nrpn[35], 0, 4095, 0, 26));
    setLEDRingPan(26, map(nrpn[32], 0, 4095, -13, 13), COLOSC1, COLOSC1, BLACK);
    //setLEDRingLine(27, map(nrpn[34], 0, 4095, 0, 26), 0x160800);
    //setLEDRingLine(28, map(nrpn[0], 0, 4095, 0, 26), 0x160800);

    // MIXER - drive
    setLEDRingVU(29, map(nrpn[15], 0, 4095, 0, 27));
    setLEDRingVU(30, map(nrpn[25], 0, 4095, 0, 27));
    setLEDRingVU(31, map(nrpn[35], 0, 4095, 0, 27));

    // FILTER RESONC / DRIVE / FENV S/R

    setLEDRingLine(33, map(nrpn[65], 0, 4095, 0, 26), COLFIL1);
    //setLEDRingVU(34, map(nrpn[90], 0, 4095, 0, 27));
    //setLEDRingLine(35, map(nrpn[62], 0, 4095, 0, 26), 0x160800);
    //setLEDRingLine(36, map(nrpn[63], 0, 4095, 0, 26), 0x160800);

    // mod env
    setLEDRingFadeIn(49, map(nrpn[75], 0, 4095, 0, 26), COLMENV);
    setLEDRingFadeOut(50, map(nrpn[76], 0, 4095, 0, 26), COLMENV);
    setLEDRingDot(51, map(nrpn[77], 0, 4095, 0, 26), COLMENV);
    setLEDRingFadeOut(52, map(nrpn[78], 0, 4095, 0, 26), COLMENV);

    // DRIVES 2

    setLEDRingVULine(57, map(nrpn[92], 0, 4095, 0, 26));
    setLEDRingVULine(58, map(nrpn[93], 0, 4095, 0, 26));

    // EQ2
    setLEDRingDot(59, map(nrpn[113], 0, 4095, 0, 26), COLEQ2);
    setLEDRingPan(60, map(nrpn[114], 0, 4095, -13, 13), COLEQ2, COLEQ2, COLEQ2);

    // MFX 2
    setLEDRingSegments(61, nrpn[105] % 26, 15, COLMFX2);
    setLEDRingDot(62, map(nrpn[106], 0, 4095, 0, 26), COLMFX2);
    setLEDRingDot(63, map(nrpn[107], 0, 4095, 0, 26), COLMFX2);
    setLEDRingDot(64, map(nrpn[108], 0, 4095, 0, 26), COLMFX2);

    for (int i = 49; i < 53; i++) setLEDBlink((i * 32) - 1, 100, PINK);
    for (int i = 0; i < 32; i++) setLEDBlink((shiftdot[i] * 32) - 1, 100, RED);

  }

  //////////////// SHIFT 2 LFO1

  else if (nrpn[150] == 2) {
    modcolor = palettes[2][1];
    // LFO 1 MAPPING
    setLEDRingQ(6, map(nrpn[351], 0, 4095, 0, 26), palettes[1][1], palettes[1][0], palettes[1][3]);
    //OSC 1
    //setLEDRingLine(9, map(nrpn[231], 0, 4095, 0, 26), GREEN);
    setLEDRingLine(10, map(nrpn[211], 0, 4095, 0, 26), modcolor);
    setLEDRingLine(11, map(nrpn[221], 0, 4095, 0, 26), modcolor);
    setLEDRingLine(12, map(nrpn[231], 0, 4095, 0, 26), modcolor);
    //OSC 2
    //setLEDRingLine(17, map(nrpn[271], 0, 4095, 0, 26), GREEN);
    setLEDRingLine(18, map(nrpn[251], 0, 4095, 0, 26), modcolor);
    setLEDRingLine(19, map(nrpn[261], 0, 4095, 0, 26), modcolor);
    setLEDRingLine(20, map(nrpn[271], 0, 4095, 0, 26), modcolor);
    //OSC 3
    //setLEDRingLine(25, map(nrpn[311], 0, 4095, 0, 26), GREEN);
    setLEDRingLine(26, map(nrpn[291], 0, 4095, 0, 26), modcolor);
    setLEDRingLine(27, map(nrpn[301], 0, 4095, 0, 26), modcolor);
    setLEDRingLine(28, map(nrpn[311], 0, 4095, 0, 26), modcolor);

    //MIXER
    setLEDRingLine(29, map(nrpn[241], 0, 4095, 0, 26), modcolor);
    setLEDRingLine(30, map(nrpn[281], 0, 4095, 0, 26), modcolor);
    setLEDRingLine(31, map(nrpn[321], 0, 4095, 0, 26), modcolor);

    //FILTER
    setLEDRingLine(33, map(nrpn[331], 0, 4095, 0, 26), modcolor);


    //AUX / DRIVE
    setLEDRingLine(57, map(nrpn[361], 0, 4095, 0, 26), modcolor);
    setLEDRingLine(58, map(nrpn[371], 0, 4095, 0, 26), modcolor);
    setLEDRingLine(59, map(nrpn[381], 0, 4095, 0, 26), modcolor);
    setLEDRingLine(60, map(nrpn[391], 0, 4095, 0, 26), modcolor);




    setLEDModDot(modcolor);
  }


  //////////////// SHIFT 3 LFO2


  else if (nrpn[150] == 3) {
    modcolor = YELLOW;
    // LFO 2 MAPPING
    setLEDRingLine(6, map(nrpn[342], 0, 4095, 0, 26), modcolor);

    //OSC 1
    //setLEDRingLine(9, map(nrpn[232], 0, 4095, 0, 26), YELLOW);
    setLEDRingLine(10, map(nrpn[212], 0, 4095, 0, 26), modcolor);
    setLEDRingLine(11, map(nrpn[222], 0, 4095, 0, 26), modcolor);
    setLEDRingLine(12, map(nrpn[232], 0, 4095, 0, 26), modcolor);
    //OSC 2
    //setLEDRingLine(17, map(nrpn[272], 0, 4095, 0, 26), YELLOW);
    setLEDRingLine(18, map(nrpn[252], 0, 4095, 0, 26), modcolor);
    setLEDRingLine(19, map(nrpn[262], 0, 4095, 0, 26), modcolor);
    setLEDRingLine(20, map(nrpn[272], 0, 4095, 0, 26), modcolor);
    //OSC 3
    //setLEDRingLine(25, map(nrpn[312], 0, 4095, 0, 26), YELLOW);
    setLEDRingLine(26, map(nrpn[292], 0, 4095, 0, 26), modcolor);
    setLEDRingLine(27, map(nrpn[302], 0, 4095, 0, 26), modcolor);
    setLEDRingLine(28, map(nrpn[312], 0, 4095, 0, 26), modcolor);

    //MIXER
    setLEDRingLine(29, map(nrpn[242], 0, 4095, 0, 26), modcolor);
    setLEDRingLine(30, map(nrpn[282], 0, 4095, 0, 26), modcolor);
    setLEDRingLine(31, map(nrpn[322], 0, 4095, 0, 26), modcolor);

    //FILTER
    setLEDRingLine(33, map(nrpn[332], 0, 4095, 0, 26), modcolor);

    //AUX / DRIVE
    setLEDRingLine(57, map(nrpn[362], 0, 4095, 0, 26), modcolor);
    setLEDRingLine(58, map(nrpn[372], 0, 4095, 0, 26), modcolor);
    setLEDRingLine(59, map(nrpn[382], 0, 4095, 0, 26), modcolor);
    setLEDRingLine(60, map(nrpn[392], 0, 4095, 0, 26), modcolor);

    setLEDModDot(modcolor);

  }


  //////////////// SHIFT 4 MOD ENV


  else if (nrpn[150] == 4) {
    modcolor = palettes[1][1];
    modcolor2 = palettes[1][0];
    modcolor3 = palettes[1][2];
    // LFO 2 MAPPING
    setLEDRingPan(6, map(nrpn[343], 0, 4095, 0, 26), modcolor, modcolor2, modcolor3);

    //OSC 1
    //setLEDRingLine(9, map(nrpn[233], 0, 4095, 0, 26), YELLOW);
    setLEDRingPan(10, map(nrpn[213], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    setLEDRingPan(11, map(nrpn[223], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    setLEDRingPan(12, map(nrpn[233], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    //OSC 2
    //setLEDRingLine(17, map(nrpn[272], 0, 4095, 0, 26), YELLOW);
    setLEDRingPan(18, map(nrpn[253], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    setLEDRingPan(19, map(nrpn[263], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    setLEDRingPan(20, map(nrpn[273], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    //OSC 3
    //setLEDRingLine(25, map(nrpn[312], 0, 4095, 0, 26), YELLOW);
    setLEDRingPan(26, map(nrpn[293], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    setLEDRingPan(27, map(nrpn[303], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    setLEDRingPan(28, map(nrpn[313], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);

    //MIXER
    setLEDRingPan(29, map(nrpn[243], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    setLEDRingPan(30, map(nrpn[283], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    setLEDRingPan(31, map(nrpn[323], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);

    //FILTER
    setLEDRingPan(33, map(nrpn[333], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);

    //AUX / DRIVE
    setLEDRingPan(57, map(nrpn[363], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    setLEDRingPan(58, map(nrpn[373], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    setLEDRingPan(59, map(nrpn[383], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    setLEDRingPan(60, map(nrpn[393], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);

    setLEDModDot(modcolor);

  }


  //////////////// SHIFT 5 MOD WHL


  else if (nrpn[150] == 5) {
    modcolor = palettes[1][1];
    modcolor2 = palettes[1][0];
    modcolor3 = palettes[1][2];
    // LFO 2 MAPPING
    setLEDRingPan(6, map(nrpn[344], 0, 4095, 0, 26), modcolor, modcolor2, modcolor3);

    //OSC 1
    //setLEDRingLine(9, map(nrpn[233], 0, 4095, 0, 26), YELLOW);
    setLEDRingPan(10, map(nrpn[214], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    setLEDRingPan(11, map(nrpn[224], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    setLEDRingPan(12, map(nrpn[234], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    //OSC 2
    //setLEDRingLine(17, map(nrpn[272], 0, 4095, 0, 26), YELLOW);
    setLEDRingPan(18, map(nrpn[254], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    setLEDRingPan(19, map(nrpn[264], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    setLEDRingPan(20, map(nrpn[274], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    //OSC 3
    //setLEDRingLine(25, map(nrpn[312], 0, 4095, 0, 26), YELLOW);
    setLEDRingPan(26, map(nrpn[294], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    setLEDRingPan(27, map(nrpn[304], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    setLEDRingPan(28, map(nrpn[314], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);

    //MIXER
    setLEDRingPan(29, map(nrpn[244], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    setLEDRingPan(30, map(nrpn[284], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    setLEDRingPan(31, map(nrpn[324], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);

    //FILTER
    setLEDRingPan(33, map(nrpn[334], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);

    //AUX / DRIVE
    setLEDRingPan(57, map(nrpn[364], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    setLEDRingPan(58, map(nrpn[374], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    setLEDRingPan(59, map(nrpn[384], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);
    setLEDRingPan(60, map(nrpn[394], 0, 4095, -13, 13), modcolor, modcolor2, modcolor3);

    setLEDModDot(modcolor);

  }



  //////////////// SHIFT 6 VELOCITY


  else if (nrpn[150] == 6) {
    // LFO 2 MAPPING
    setLEDRingLine(6, map(nrpn[345], 0, 4095, 0, 26), BLUE3);

    //OSC 1
    //setLEDRingLine(9, map(nrpn[233], 0, 4095, 0, 26), YELLOW);
    setLEDRingLine(10, map(nrpn[215], 0, 4095, 0, 26), BLUE3);
    setLEDRingLine(11, map(nrpn[225], 0, 4095, 0, 26), BLUE3);
    setLEDRingLine(12, map(nrpn[235], 0, 4095, 0, 26), BLUE3);
    //OSC 2
    //setLEDRingLine(17, map(nrpn[272], 0, 4095, 0, 26), YELLOW);
    setLEDRingLine(18, map(nrpn[255], 0, 4095, 0, 26), BLUE3);
    setLEDRingLine(19, map(nrpn[265], 0, 4095, 0, 26), BLUE3);
    setLEDRingLine(20, map(nrpn[275], 0, 4095, 0, 26), BLUE3);
    //OSC 3
    //setLEDRingLine(25, map(nrpn[312], 0, 4095, 0, 26), YELLOW);
    setLEDRingLine(26, map(nrpn[295], 0, 4095, 0, 26), BLUE3);
    setLEDRingLine(27, map(nrpn[305], 0, 4095, 0, 26), BLUE3);
    setLEDRingLine(28, map(nrpn[315], 0, 4095, 0, 26), BLUE3);

    //MIXER
    setLEDRingLine(29, map(nrpn[245], 0, 4095, 0, 26), BLUE3);
    setLEDRingLine(30, map(nrpn[285], 0, 4095, 0, 26), BLUE3);
    setLEDRingLine(31, map(nrpn[325], 0, 4095, 0, 26), BLUE3);

    //FILTER
    setLEDRingLine(33, map(nrpn[335], 0, 4095, 0, 26), BLUE3);

    //AUX / DRIVE
    setLEDRingLine(57, map(nrpn[365], 0, 4095, 0, 26), BLUE3);
    setLEDRingLine(58, map(nrpn[375], 0, 4095, 0, 26), BLUE3);
    setLEDRingLine(59, map(nrpn[385], 0, 4095, 0, 26), BLUE3);
    setLEDRingLine(60, map(nrpn[395], 0, 4095, 0, 26), BLUE3);

    setLEDModDot(BLUE3);

  }

  //////////////// SHIFT 7 AT


  else if (nrpn[150] == 7) {
    // LFO 2 MAPPING
    setLEDRingLine(6, map(nrpn[346], 0, 4095, 0, 26), BLUE4);

    //OSC 1
    //setLEDRingLine(9, map(nrpn[233], 0, 4095, 0, 26), YELLOW);
    setLEDRingLine(10, map(nrpn[216], 0, 4095, 0, 26), BLUE4);
    setLEDRingLine(11, map(nrpn[226], 0, 4095, 0, 26), BLUE4);
    setLEDRingLine(12, map(nrpn[236], 0, 4095, 0, 26), BLUE4);
    //OSC 2
    //setLEDRingLine(17, map(nrpn[272], 0, 4095, 0, 26), YELLOW);
    setLEDRingLine(18, map(nrpn[256], 0, 4095, 0, 26), BLUE4);
    setLEDRingLine(19, map(nrpn[266], 0, 4095, 0, 26), BLUE4);
    setLEDRingLine(20, map(nrpn[276], 0, 4095, 0, 26), BLUE4);
    //OSC 3
    //setLEDRingLine(25, map(nrpn[312], 0, 4095, 0, 26), YELLOW);
    setLEDRingLine(26, map(nrpn[296], 0, 4095, 0, 26), BLUE4);
    setLEDRingLine(27, map(nrpn[306], 0, 4095, 0, 26), BLUE4);
    setLEDRingLine(28, map(nrpn[316], 0, 4095, 0, 26), BLUE4);

    //MIXER
    setLEDRingLine(29, map(nrpn[246], 0, 4095, 0, 26), BLUE4);
    setLEDRingLine(30, map(nrpn[286], 0, 4095, 0, 26), BLUE4);
    setLEDRingLine(31, map(nrpn[326], 0, 4095, 0, 26), BLUE4);

    //FILTER
    setLEDRingLine(33, map(nrpn[336], 0, 4095, 0, 26), BLUE4);

    //AUX / DRIVE
    setLEDRingLine(57, map(nrpn[366], 0, 4095, 0, 26), BLUE4);
    setLEDRingLine(58, map(nrpn[376], 0, 4095, 0, 26), BLUE4);
    setLEDRingLine(59, map(nrpn[386], 0, 4095, 0, 26), BLUE4);
    setLEDRingLine(60, map(nrpn[396], 0, 4095, 0, 26), BLUE4);

    setLEDModDot(BLUE3);

  }
}



//////////// FUNCTIONS
void setModdot() {
  moddot[0] = 6;

  moddot[1] = 10;
  moddot[2] = 11;
  moddot[3] = 12;

  moddot[4] = 18;
  moddot[5] = 19;
  moddot[6] = 20;

  moddot[7] = 26;
  moddot[8] = 27;
  moddot[9] = 28;

  moddot[10] = 29;
  moddot[11] = 30;
  moddot[12] = 31;

  moddot[13] = 33;

  moddot[14] = 57;
  moddot[15] = 58;
  moddot[16] = 59;
  moddot[17] = 60;

  moddotnrpn[0] = 15;

  moddotnrpn[1] = 1;
  moddotnrpn[2] = 2;
  moddotnrpn[3] = 3;

  moddotnrpn[4] = 5;
  moddotnrpn[5] = 6;
  moddotnrpn[6] = 7;

  moddotnrpn[7] = 9;
  moddotnrpn[8] = 10;
  moddotnrpn[9] = 11;

  moddotnrpn[10] = 4;
  moddotnrpn[11] = 8;
  moddotnrpn[12] = 12;

  moddotnrpn[13] = 13;

  moddotnrpn[14] = 16;
  moddotnrpn[15] = 17;
  moddotnrpn[16] = 18;
  moddotnrpn[17] = 19;
}

void setRingMatrix() {
  ringMatrix[0][0] = 1;
  ringMatrix[0][1] = 2;
  ringMatrix[0][2] = 3;
  ringMatrix[0][3] = 4;

  ringMatrix[0][4] = 9;
  ringMatrix[0][5] = 10;
  ringMatrix[0][6] = 11;
  ringMatrix[0][7] = 12;

  ringMatrix[0][8] = 29;

  ringMatrix[0][9] = 0;
  ringMatrix[0][10] = 33;
  ringMatrix[0][11] = 34;
  ringMatrix[0][12] = 0;

  ringMatrix[0][13] = 0;
  ringMatrix[0][14] = 0;
  ringMatrix[0][15] = 0;
  ringMatrix[0][16] = 0;

  ///

  ringMatrix[1][0] = 0;
  ringMatrix[1][1] = 0;
  ringMatrix[1][2] = 0;
  ringMatrix[1][3] = 0;

  ringMatrix[1][4] = 17;
  ringMatrix[1][5] = 18;
  ringMatrix[1][6] = 19;
  ringMatrix[1][7] = 20;

  ringMatrix[1][8] = 30;

  ringMatrix[1][9] = 41;
  ringMatrix[1][10] = 42;
  ringMatrix[1][11] = 43;
  ringMatrix[1][12] = 44;

  ringMatrix[1][13] = 57;
  ringMatrix[1][14] = 58;
  ringMatrix[1][15] = 59;
  ringMatrix[1][16] = 60;

  ///

  ringMatrix[2][0] = 5;
  ringMatrix[2][1] = 6;
  ringMatrix[2][2] = 7;
  ringMatrix[2][3] = 8;

  ringMatrix[2][4] = 25;
  ringMatrix[2][5] = 26;
  ringMatrix[2][6] = 27;
  ringMatrix[2][7] = 28;

  ringMatrix[2][8] = 31;

  ringMatrix[2][9] = 49;
  ringMatrix[2][10] = 50;
  ringMatrix[2][11] = 51;
  ringMatrix[2][12] = 52;

  ringMatrix[2][13] = 61;
  ringMatrix[2][14] = 62;
  ringMatrix[2][15] = 63;
  ringMatrix[2][16] = 64;
}



void setShiftdot() {
  shiftdot[0] = 5;
  shiftdot[1] = 6;
  shiftdot[2] = 10;
  shiftdot[3] = 18;
  shiftdot[4] = 26;
  shiftdot[5] = 29;
  shiftdot[6] = 30;
  shiftdot[7] = 31;
  shiftdot[8] = 33;
  shiftdot[9] = 57;
  shiftdot[10] = 58;
  shiftdot[11] = 59;
  shiftdot[12] = 60;
  shiftdot[13] = 61;
  shiftdot[14] = 62;
  shiftdot[15] = 63;
}

void setPalettes() {
  for (int i = 0; i < 4; i++) {
    palettes[0][i] = RED;
    palettes[1][i] = BLUE;
    palettes[2][i] = GREEN;

    palettes[3][i] = YELLOW;
    palettes[4][i] = ORANGE;
    palettes[5][i] = PINK;
    palettes[6][i] = WHITE;
    palettes[7][i] = CRGB::Aqua;
    palettes[8][i] = CRGB::Aquamarine;
    palettes[9][i] = CRGB::CadetBlue;
    palettes[10][i] = CRGB::CRGB::Coral;
    palettes[11][i] = CRGB::Aquamarine;
    palettes[12][i] = CRGB::Aquamarine;

  }


  for (int i = 0; i < 16; i++) {
    palettes[i][0].addToRGB(30);
    palettes[i][2].fadeLightBy(80);
    palettes[i][3].fadeLightBy(160);
  }
}

void setVUPalette() {
  VUPalette[0] = CRGB(0, 100, 0);
  VUPalette[1] = CRGB(0, 170, 0);
  VUPalette[2] = CRGB(0, 255, 0);
  VUPalette[3] = CRGB(200, 255, 0);
  VUPalette[4] = CRGB::Yellow;
  VUPalette[5] = CRGB(255, 150, 0);
  VUPalette[6] = CRGB(255, 0, 0);
  VUPalette[7] = CRGB(255, 0, 0);
}

void setcyclecolor() {
  cyclecolor[0] = 0xFF0000;
  cyclecolor[1] = 0xFF8f00;
  cyclecolor[2] = 0x0000FF;
  cyclecolor[3] = 0x00ff00;

}
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////// LED RING FUNCTIONS
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void clearLEDRing(int ringNr) {
  int offseti = (ringNr - 1) * 32;
  for (int i = 0; i < 32; i++) leds[offseti + i] = 0;
}

void clearLEDRingAll(void) {
  for (int i = 0; i < NUM_PIXELS; i++) leds[i] = 0;
}


// Sets one LED at VALUE to COLOR
void setLEDRingDot(int ringNr, int value, CRGB color) {
  leds[(ringNr - 1) * 32 + 2 + value] = color;
}

// Sets all LEDs up to VALUE to COLOR
void setLEDRingLine(int ringNr, int value, CRGB color) {
  int offseti = (ringNr - 1) * 32 + 2;
  for (int i = 0; i <= value; i++) {
    leds[offseti + i] = color;
  }
}


// VU Style green-yellow-red
void setLEDRingVU(int ringNr, int value) {
  int offseti = (ringNr - 1) * 32 + 2;
  for (int i = 0; i <= value; i++) {
    leds[offseti + i] = ColorFromPalette(VUPalette, i * 4, 255, LINEARBLEND);
  }
}


// VU Style green-yellow-red line
void setLEDRingVULine(int ringNr, int value) {
  int offseti = (ringNr - 1) * 32 + 2;
  for (int i = 0; i <= value; i++) {
    leds[offseti + i] = ColorFromPalette(VUPalette, value * 4, 255, LINEARBLEND);
  }
}

// Divides LED Ring into SEGMENTS
void setLEDRingDivide(int ringNr, int segments, CRGB color, CRGB bgcolor) {
  int stepSize = 31 / segments;
  int offseti = (ringNr - 1) * 32 + 2;
  for (int i = 0; i < 32; i++) leds[offseti + i] = bgcolor;
  for (int i = 0; i < segments + 1; i++) leds[offseti + (i * stepSize)] = color;
}


// Multiply LED Ring by STEPSIZE
void setLEDRingMultiply(int ringNr, int stepSize, CRGB color, CRGB bgcolor) {
  if (stepSize == 0) stepSize = 32;
  int offseti = (ringNr - 1) * 32 + 2;
  for (int i = 0; i < 32; i++) leds[offseti + i] = bgcolor;
  for (int i = 0; i < 32; i = i + stepSize) leds[offseti + i] = color;
}


// Multiply LED Ring by STEPSIZE, Fading out
void setLEDRingMultiplyFade(int ringNr, int stepSize, CRGB color, CRGB bgcolor) {
  if (stepSize == 0) stepSize = 32;
  int offseti = (ringNr - 1) * 32 + 2;
  for (int i = 0; i < 32; i++) leds[offseti + i] = bgcolor;
  for (int i = stepSize; i < 32; i = i + stepSize) {
    leds[offseti + i] = color;
    leds[offseti + i] = (i * 8);
  }
  leds[offseti] = color;
}


void setLEDRingFadeIn(int ringNr, int value, CRGB color) {
  int offseti = (ringNr - 1) * 32 + 2;
  int j = 255 / value;
  for (int i = 0; i <= value; i++) {
    leds[offseti + i] = CRGB(
                          map(color.r, 0 , 255, 0, i * j),
                          map(color.g, 0 , 255, 0, i * j),
                          map(color.b, 0 , 255, 0, i * j)

                        );
  }
  leds[offseti + value] = color;
}

void setLEDRingFadeOut(int ringNr, int value, CRGB color) {
  int offseti = (ringNr - 1) * 32 + 2;
  int j = 255 / value;
  for (int i = 0; i <= value; i++) {
    leds[offseti + i] = CRGB(
                          map(color.r, 0 , 255, 0, 255 - (i * j)),
                          map(color.g, 0 , 255, 0, 255 - (i * j)),
                          map(color.b, 0 , 255, 0, 255 - (i * j))
                        );
  }
}

// block segment number VALUE out of SEGMENTS
void setLEDRingSegments(int ringNr, int value, int segments, CRGB color) {
  int stepSize = 216 / segments;
  int offseti = (ringNr - 1) * 32 + 2 + ((value * stepSize) >> 3);
  for (int i = 0; i <= (stepSize >> 3); i++) {
    leds[offseti + i] = color;
  }
}

////////////////////////////////////////////////////////////////////////////////////
/////////////////////////PANNING / BIPOLAR///////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////


// Panning v2
void setLEDRingPan(int ringNr, int value, CRGB valcolor, CRGB centercolor, CRGB linecolor) {
  int offseti = (ringNr - 1) * 32 + 15;
  if (value > 0) {
    for (int i = 1; i < value; i++) leds[offseti + i] = linecolor;
  }
  else if (value < 0) {
    for (int i = -1; i > value ; i--) leds[offseti + i] = linecolor;
  }
  leds[offseti + value] = valcolor;
  leds[offseti] = centercolor;
}

void setLEDRingPitch(int ringNr, int value, CRGB valcolor, CRGB centercolor) {
  int offseti = (ringNr - 1) * 32 + 15;
  leds[offseti] = centercolor;
  if (abs(value) > 24) {
    if (value > 0) leds[offseti + value - 24] = valcolor.subtractFromRGB(180);
    else leds[offseti + value + 24] = valcolor.subtractFromRGB(180);
  }
  else if (abs(value) > 12) {
    if (value > 0) leds[offseti + value - 12] = valcolor.subtractFromRGB(100);
    else leds[offseti + value + 12] = valcolor.subtractFromRGB(100);
  }
  else leds[offseti + value] = valcolor;
}

// QPanning
void setLEDRingQ(int ringNr, int value, CRGB valcolor, CRGB centercolor, CRGB linecolor) {
  int offseti = (ringNr - 1) * 32 + 15;
  for (int i = 1; i <= value; i++) {
    leds[offseti + i] = linecolor;
    leds[offseti - i] = linecolor;
  }
  leds[offseti + value] = valcolor;
  leds[offseti - value] = valcolor;
  leds[offseti] = centercolor;
}

////////////////////////////////////////////////////////////////////////////////////
/////////////////////////SINGLE LED///////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////


// Blink one LED at LEDNR to COLOR at BPM speed
void setLEDBlink(int ledNr, int bpm, CRGB color) {

  int period = 60000 / bpm;
  static unsigned int prevMil;

  if (millis() >= (prevMil + period)) {
    prevMil = millis();
  }

  else if (millis() >= (prevMil + 90)) {
    leds[ledNr] = BLACK;
  }

  else {
    leds[ledNr] = color;
  }
}


// Pulse one LED at LEDNR to COLOR at BPM speed
void setLEDPulse(int ledNr, int bpm, CRGB color) {
  int period = 60000 / bpm;
  static CRGB col2;
  static unsigned int prevMil2;

  if (millis() >= (prevMil2 + period)) {
    prevMil2 = millis();
    col2 = color;

  }

  col2.subtractFromRGB(4);
  leds[ledNr] = col2;

}


void setLEDVLine (int value, CRGB color) {
  //16x17horiz pixel = 272
  int ringHoriz = value >> 4;
  for (int i = 0; i <= 2; i++) {
    setLEDRingHLine(ringMatrix[i][ringHoriz], value % 16, color);
  }
}

void setLEDHLine (int value, CRGB color) {
  //16x17horiz pixel = 272
  int ringVert = value >> 4;
  for (int i = 0; i < 17; i++) {
    setLEDRingVLine(ringMatrix[ringVert][i], value % 16, color);
  }
}


void setLEDHRGBLine (int value, CRGB color) {
  //16x17horiz pixel = 272
  int ringVert = value >> 4;
  for (int i = 0; i < 17; i++) {
    setLEDRingVLine(ringMatrix[ringVert][i], value % 16, CHSV(i << 4, 255, 255));
  }
}

void setLEDRingNightrider(int ringNr, CRGB color) {
  int offseti = (ringNr - 1) * 32 + 2;
  static bool dir;
  if (bpmGlobalTrigger) {
    leds[offseti + bpmGlobal27] = color;
    leds[offseti + bpmGlobal27 - 1] = color.nscale8( 192);
    leds[offseti + bpmGlobal27 - 2] = color.nscale8( 128);
    leds[offseti + bpmGlobal27 - 3] = color.nscale8( 96);
    leds[offseti + bpmGlobal27 - 4] = color.nscale8( 64);
    leds[offseti + bpmGlobal27 - 5] = color.nscale8( 32);
    leds[offseti + bpmGlobal27 - 6] = color.nscale8( 16);
  }
  else {
    leds[offseti - bpmGlobal27 + 27] = color;
    leds[offseti - bpmGlobal27 + 28] = color.nscale8( 192);
    leds[offseti - bpmGlobal27 + 29] = color.nscale8( 128);
    leds[offseti - bpmGlobal27 + 30] = color.nscale8( 96);
    leds[offseti - bpmGlobal27 + 31] = color.nscale8( 64);
    leds[offseti - bpmGlobal27 + 32] = color.nscale8( 32);
    leds[offseti - bpmGlobal27 + 33] = color.nscale8( 16);
  }
  if (bpmGlobal27 == 0) dir = !dir;
  /*
    Serial.print(dir);
    Serial.print(" .  ");
    Serial.print(bpmGlobal27);
    Serial.print(" .  ");
    Serial.print(bpmGlobalTrigger);
    Serial.println(" .  ");
  */
}



////////////////////////////////////////////////////////////////////////////////////
/////////////////////////HELPER / OTHERS///////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////


// swipe dot
void setLEDRingSwipe(int ringNr, CRGB color) {
  int offseti = (ringNr - 1) * 32 + 2;
  leds[offseti + (runner % 30)] = color;
}

void setLEDRingSwipeSlow(int ringNr, CRGB color) {
  int offseti = (ringNr - 1) * 32 + 2;
  leds[offseti + ((runner >> 1) % 30)] = color;
}
void setLEDRingSwipeSlow2(int ringNr, CRGB color) {
  int offseti = (ringNr - 1) * 32 + 2;
  leds[offseti + ((runner >> 2) % 30)] = color;
}
void setLEDRingSwipeSlow4(int ringNr, CRGB color) {
  int offseti = (ringNr - 1) * 32 + 2;
  leds[offseti + ((runner >> 3) % 30)] = color;
}

void setLEDRingBlinkFull(int ringNr, CRGB color) {
  int offseti = (ringNr - 1) * 32 + 2;
  if ((runner % 16) < 8) {
    for (int i = 0; i < 27; i++) {
      leds[offseti + i] = color;
    }
  } else {
    for (int i = 0; i < 27; i++) {
      leds[offseti + i] = 0x000000;
    }
  }
}


void setLEDModDot(CRGB color) {
  for (int i = 0; i < 31; i++) {
    leds[(moddot[i] * 32) - 1] = color;
  }
}



void setLEDRingVLine(int ringNr, int value, CRGB color) {
  int offseti = (ringNr * 32) - 17;
  leds[offseti + value] += color;
  leds[offseti - value] += color;
}

void setLEDRingHRand(int ringNr, int value, CRGB color) {
  int offseti = (ringNr * 32) - 17;
  bool odd = random(2);
  //Serial.println(odd);

  if (odd) leds[offseti + value] = color;
  else leds[offseti - value] = color;
}

void setLEDRingHLine(int ringNr, int value, CRGB color) {
  int offseti = (ringNr * 32) - 25;
  if (value < 8) {
    leds[offseti + value] += color;
    leds[offseti - value] += color;
  }
  else {
    leds[offseti + value] += color;
    leds[offseti + 32 - value] += color;
  }

}



////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////SCREENSAVER////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////

void saverSnowfall(int spd) {
  while (!MIDI.read()) {
    int l = random(255);
    if ((saveCount % (spd > 1)) == 0) {
      if ((saveCount >> (spd - 3)) <= 17) setLEDRingHRand(ringMatrix[0][random(17)], random((saveCount >> (spd - 3)) % 17), CRGB(l, l, l));
      else if ((saveCount >> (spd - 3)) <= 34) {
        setLEDRingHRand(ringMatrix[0][random(17)], random(17), CRGB(l, l, l));
        setLEDRingHRand(ringMatrix[1][random(17)], random((saveCount >> (spd - 3)) % 17), CRGB(l, l, l));
      }
      else if ((saveCount >> (spd - 3)) > 34) {
        setLEDRingHRand(ringMatrix[0][random(17)], random(17), CRGB(l, l, l));
        setLEDRingHRand(ringMatrix[1][random(17)], random(17), CRGB(l, l, l));
        setLEDRingHRand(ringMatrix[2][random(17)], random((saveCount >> (spd - 3)) % 17), CRGB(l, l, l));
      }
    }
    for (int i = 0; i < 2047; i++) leds[i].fadeToBlackBy(1);
    delay(spd << 1);
    LEDS.show();
    saveCount++;
    if (millis() > (saveMillis + saveChangeTime)) {
      pickedMode = 0;
      break;
    }
  }
  if (pickedMode) {
    //saveCount = saveTime + 2;  //test & copy to all
    pickedMode = 0;
  }
}

void saverAddRandomBWFade(int spd) {
  while (!MIDI.read()) {
    int l = random (255);
    leds[random(2048)] += CRGB (l, l, l);
    for (int i = 0; i < 2047; i++) leds[i].fadeToBlackBy(1);
    delay(spd);
    LEDS.show();
    if (millis() > (saveMillis + saveChangeTime)) {
      pickedMode = 0;
      break;
    }
  }
  pickedMode = 0;
}

void saverAddRandom(int spd) {
  while (!MIDI.read()) {
    leds[random(2048)] += CHSV(random(255), random(255), random(64));
    delay(spd);
    LEDS.show();
    if (millis() > (saveMillis + saveChangeTime)) {
      pickedMode = 0;
      break;
    }
  }
  pickedMode = 0;
}


/////////////// non interrupt
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void saverHWave(int spd, int lngth, CRGB color) {
  int dir = (((saveCount << spd) >> 2) % ((17 + lngth) * 8)) >> 3;
  //Serial.println(dir);
  for (int j = 0; j < 3; j++) {
    for (int k = 0; k < lngth; k++) {
      CRGB c2 = color;
      if (((dir - k) > 0) && ((dir - k) < 17)) setLEDRingLine(ringMatrix[j][dir - k], 32, c2.subtractFromRGB((k * lngth * 4)));
    }
  }
}



void saverHueSnake (int spd) {
  int dir = ((saveCount << spd) % 2300);
  for (int i = 0; i < 255; i++) {
    int k = 2048 - dir + i;
    if ((k > 0) && (k < 2048)) leds[k] += CHSV(dir % 255, 255, i);
  }
}

void saverNightriderFine (int spd, int nwidth, CRGB color) {

  int dir = (((saveCount << spd) >> 2) % 544);
  if (dir > 272)dir = 544 - dir;
  for (int i = 0; i < nwidth; i++) {
    if ((dir + i) < 272) setLEDVLine(dir + i, color);   // needs >=0 check
    setLEDVLine(dir - i, color);
  }

}

void saverNightriderFineH (int spd, int nwidth, CRGB color) {

  int dir = (((saveCount << spd) >> 2) % 93);
  if (dir > 47) dir = 93 - dir;
  for (int i = 0; i < nwidth; i++) {
    if ((dir + i) < 47) setLEDHLine(dir + i, color);    // needs >=0 check
    setLEDHLine(dir, color);
  }
}

void saverNightriderFineHRGB (int spd, int nwidth, CRGB color) {

  int dir = (((saveCount << spd) >> 2) % 93);
  if (dir > 47) dir = 93 - dir;
  for (int i = 0; i < nwidth; i++) {
    if ((dir + i) < 47) setLEDHLine(dir + i, color);      // needs >=0 check
    setLEDHRGBLine(dir, color);
  }
}




void saverHSwipeAll(int spd) {

  int dir = ((saveCount << spd) % 128) >> 3;
  for (int i = 0; i < 17; i++) {
    for (int j = 0; j <= 2; j++) setLEDRingHLine(ringMatrix[j][i], dir, RED);
  }
}

void saverVSwipeAll (int spd) {
  int dir = ((saveCount << spd) % 384) >> 3;

  if (dir < 16) {
    for (int i = 0; i < 17; i++) {
      setLEDRingVLine(ringMatrix[0][i], dir, RED);
    }
  }
  else if (dir < 32) {
    for (int i = 0; i < 17; i++) {
      setLEDRingVLine(ringMatrix[1][i], dir - 16, RED);
    }
  }
  else if (dir < 48) {
    for (int i = 0; i < 17; i++) {
      setLEDRingVLine(ringMatrix[2][i], dir - 32, RED);
    }
  }
}

void saverNightriderCoarse (int spd, CRGB color) {


  int dir = ((saveCount << spd) % 256) >> 3;
  if (dir >= 17) dir = 32 - dir;
  for (int i = 0; i <= 2; i++) {
    setLEDRingLine(ringMatrix[i][dir], 27, color);

  }

  delay(spd + 1);
}
/*
  void saverNightriderCoarse (int spd, CRGB color) {


  int dir = ((saveCount << spd) % 256) >> 3;
  if (dir >= 17) dir = 32 - dir;
  if (dir < 17) && ((dir + k) >= 0)) setLEDRingLine(ringMatrix[i][dir + k], 27, col );
  if (dir < 17) {
    for (int i = 0; i <= 2; i++) {
      setLEDRingLine(ringMatrix[i][dir], 27, color);

    }
  }
  else {
    for (int i = 0; i <= 2; i++) {
      setLEDRingLine(ringMatrix[i][33 - dir], 27, RED);
    }
  }
  delay(spd + 1);
  }
*/

void saverNightriderCoarseFade(int spd, CRGB color) {


  int dir = ((saveCount << spd) % 272) >> 3;
  if (dir >= 18) dir = 34 - dir;
  for (int i = 0; i <= 2; i++) {
    for (int k = -2; k <= 2; k++) {
      CRGB col = color;
      if (abs(k) > 0) col = col / (abs(k) << 2);
      if (((dir + k) < 17) && ((dir + k) >= 0)) setLEDRingLine(ringMatrix[i][dir + k], 27, col );
      Serial.println(abs(k));
    }
  }

  delay(spd + 1);
}
